#!/usr/bin/env perl
#####
# keywords.pl
# Andy Hammerlindl 2006/07/31
#
#  Extract keywords from camp.l and list them in a keywords file.  These
#  keywords are used in autocompletion at the interactive prompt.
#####

# Extra keywords to add that aren't in camp.l.  Currently, this is just the
# special commands of the interactive prompt, which I couldn't automatically
# extract from interact.cc.
@extrawords = ("help", "quit", "exit", "input", "reset");

open(keywords, ">keywords.cc") ||
    die("Couldn't open keywords.out for writing.");

print keywords <<END;
/*****
 * This file is automatically generated by keywords.pl.
 * Changes will be overwritten.
 *****/

END

sub add {
  print keywords "ADD(".$_[0].");\n";
}

foreach $word (@extrawords) {
  add($word);
}

open(camp, "camp.l") || die("Couldn't open camp.l");

# Search for the %% separator, after which the definitions start.
while (<camp>) {
  if (/^%%\s*$/) {
    last; # Break out of the loop.
  }
}

while (<camp>) {
  if (/^%%\s*$/) {
    last; # A second %% indicates the end of definitions.
  }
  if (/^(\w+)\s*\{/) {
    add($1);
  }
}
